/* lcd.c			L. Cicchinelli		Z-World, Inc		April 2000
Copyright Z-World Inc. June 2000			Larry Cicchinelli

Product: Jackrabbit single board computer.

In response to several customer requests, the following sample program was written 
by the Z-World Technical Support Staff.  The sample code has NOT been subjected to 
standard software testing and validation procedures, and is being made available to 
users as a convenience only. There is no warranty or guarantee, implied or otherwise.

The hardware and software information below pertains to connecting an LCD which uses 
the HD44780 controller or an equivalent.  This file contains the only documentation for
this program.

This file contains information for both a 4 bit and 8 bit interface.  The 4 bit
interface uses one of the parallel I/O ports while the 8 bit interface uses the
external I/O feature of the Rabbit 2000 microprocessor.

**********************************************************************
****************************** 4 Bit Mode ****************************
**********************************************************************

Functions which are unique to the 4 bit interface are named fn4...

This is the initialization sequence for a 2 by 20 character LCD programmed in 4
bit mode (chosen in order to reduce the number of I/O pins):
	delay for power up - allow LCD to reset
	command: 4 bit mode
	command: 2 lines, 5x10 character size, and 4 bit mode again
	command: turn on display and cursor - non blinking
	command: incr address and shift cursor with each character
note: each command must be followed by a delay to allow the display to respond.

Here is the connection list: 
Jackrabbit Pin LCD Signal
	PA0 J3-3		DB4
	PA1 J3-4		DB5
	PA2 J3-5		DB6
	PA3 J3-6		DB7

	PA6 J3-9		RS		Register Select: 0=command, 1=data
	PA7 J3-10	E		normally low: latches on high to low transition

DB0 - DB3 of the LCD are grounded as is the R/W pin.  Since we are using 4 bit
mode DB0 - DB3 are not used by the controller.  We will only be writing so we
ground the R/W pin

Since Port A is used to control the LEDs on the expansion board, they will be
controlled by the data values which you program.

VEE is used to control brightness - the simplest thing to do is ground it. If
you need to control the brightness you can connect a potentiometer between
ground and -5V with the arm going to the VEE pin. Check the specs on the LCD
before doing this as some may require a different connection.  VCC is +5 and
VSS is ground. 


**********************************************************************
****************************** 8 Bit Mode ****************************
**********************************************************************

Functions which are unique to the 8 bit interface are named fn8...

This is the initialization sequence for a 2 by 20 character LCD programmed in 8
bit mode:
	delay for power up - allow LCD to reset
	command: 8 bit mode
	command: 2 lines, 5x10 character size, and 8 bit mode again
	command: turn on display and cursor - non blinking
	command: incr address and shift cursor with each character
note: each command must be followed by a delay to allow the display to respond.

Here is the connection list: 
Core Module	   LCD
 Signal		  Signal
	D0				DB0
	D1				DB1
	D2				DB2
	D3				DB3
	D4				DB4
	D5				DB5
	D6				DB6
	D7				DB7
	A0				RS		Register Select: 0=command, 1=data
	A1				R/W	0=write, 1=read
	*				E		normally low: latches on high to low transition

* = circuit which uses a 74HC02, or equivalent, as follows

			|----\
/PE0-----O		|	|-----|	|----\
			|		|--|		|	|		|
/IOR-----O		|	|		|	|		|
			|----/	|		|	|		|
						|		O--O		|---- E
			|----\	|		|	|		|
/PE0-----O		|	|		|	|		|
			|		|--|		|	|		|
/IOW-----O		|	|-----|	|----/
			|----/

	Bus address		Function
		0				write command byte
		1				write data byte
		2				read busy flag and address
		3				read data from RAM

	See notes above for power and brightness.

**********************************************************************
**********************************************************************
*********************************************************************/

/*********************************************************************
fnMsDelay - delay some number of milliseconds
	input parameter: int number of milliseconds to delay
	return value: none
	errors: none
*/
void fnMsDelay ( int iDelay )
{	unsigned long ul0;
	ul0 = MS_TIMER;				// get current timer value
	while ( MS_TIMER < ul0 + (unsigned long) iDelay );
}

/*********************************************************************
fnUsDelay - delay some number of micro seconds - very approximate!!
	about 11usec per iteration with the 7.3MHz crystal.  The formula
	used here was determined experimentally.
*/
void fnUsDelay ( int iDelay )
{	int i;
	iDelay /= 11;
	for ( i=0; i<iDelay; i++ );
}

/********************************************************************/
/**************************** 4 bit functions ***********************/
/********************************************************************/

#define DATA		0x40
#define COMMAND	0x00
int iDataFlag;

/*********************************************************************
fn4OneNib - send a single nibble to the LCD
*/
void fn4OneNib ( char cNib )
{	cNib &= 0x0F;								// remove upper 4 bits
	cNib |= iDataFlag;						// insert data flag for RS
	WrPortI ( PADR, NULL, cNib|0x80 );	// assert E
	WrPortI ( PADR, NULL, cNib );			// remove E
}


/*********************************************************************
				for those who prefer assembly code.........
#asm
fn4OneNib::
; character is in HL
			ld		a, L			; get the character
			and	0x0F			; insure upper nibble is zero
			ld		hl, iDataFlag ; get address of data flag
			or		(hl)			; insert Data/Command bit
			or		0x80			; assert E
	ioi	ld		(PADR),a		; ship data value and E
			and	0x7F			; remove E
	ioi	ld		(PADR),a		; ship data value without E
			ret
#endasm
/**/

/*********************************************************************
fn4Byte - send two nibbles to the LCD, upper nibble first
*/
void fn4Byte ( char cNib )
{	int i;
	i = cNib;				// temp store
	cNib >>= 4;				// put upper nibble into lower 4 bits
	fn4OneNib ( cNib );	// send upper nibble
	fn4OneNib ( i );		// send lower nibble
	fnUsDelay (100);
}

/*********************************************************************
fn4LCD_Init - Initialize the LCD for the following operating parameters:
	4 bit mode, 2 lines, 5x10
	turn on display and cursor: non-blinking
	incr address and shift cursor with each character
*/
void fn4LCD_Init ()
{	WrPortI ( PADR, NULL, 0 );		// write 0 to port A output register
	WrPortI ( SPCR, NULL, 0x84 );	// make port A output
	iDataFlag = COMMAND;				// show command mode
	fnMsDelay ( 1000 );				// wait for LCD to reset itself
	fn4OneNib ( 2 );					// 4 bit mode
	fnUsDelay (100);
	fn4Byte ( '\B00101100' );		// 4 bit mode, 2 lines, 5x10
	fn4Byte ( '\B00001110' );		// turn on display and cursor: non-blinkin
	fn4Byte ( '\B00000110' );		// incr address and shift cursor with each character
}


/*********************************************************************
fn4Display - display a line of text on the LCD
	argument: address of null terminated text string
*/
void fn4Display ( char *szp )
{	iDataFlag = DATA;					// next bytes are data
	while ( *szp ) fn4Byte ( *szp++ );
}

void fn4Clear ( void )
{	iDataFlag = COMMAND;				// set up for command
	fn4Byte ( 0x01 );					// clear the display
	fnMsDelay (3);						// insure at least 2 msec
}

void fn4Line2 ( void )
{	iDataFlag = COMMAND;				// set up for command
	fn4Byte ( 0xC0 );					// set RAM address to Line 2
					//note: this value may vary for different displays
	fnMsDelay (3);						// insure at least 2 msec
}

/********************************************************************/
/**************************** 8 bit functions ***********************/
/********************************************************************/

#define	WCOMMAND	0
#define	WDATA		1
#define	RBUSY		2	// not currrently used
#define	RRAM		3	// not currrently used


/*********************************************************************
fn8Byte - send a byte to the LCD
*/
void fn8ByteC ( int iAddr, char cByte )
{
	WrPortE ( iAddr, NULL, (int)cByte );
	fnUsDelay (100);
}

#asm
;function requires two parameters:
;@sp+2= 1st parameter value = external I/O address
;@sp+4= 2nd parameter value = data
;
;Need to clock doubler in order to meet the set up and hold
;time requirements of the LCD.  For a faster CPU time you may
;need to set the CPU clock to "divide by 8" mode

fn8Byte::
			xor	a				;a = 0
	ioi	ld		(GCDR), a	;turn off clock doubler

			ld		hl,(sp+4)	; get data value
			ld		a, l			; save it for ioe
			ld		hl,(sp+2)	; get data address
	ioe	ld		(hl),a		; ship data value

			ld		a, (GCDRShadow)
	ioi	ld		(GCDR), a	;restore clock doubler

			ld		hl, 100		;set up for
			push	hl
			call	fnUsDelay	;delay
			pop	hl
			ret
#endasm


/*********************************************************************
fn8LCD_Init - Initialize the LCD for the following operating parameters:
	4 bit mode, 2 lines, 5x10
	turn on display and cursor: non-blinking
	incr address and shift cursor with each character
*/
void fn8LCD_Init ()
{
			// set up PE0 as I/O strobe
	WrPortI ( PEDDR, &PEDDRShadow, 1 );	// PE0 = output
	WrPortI ( PEFR, &PEFRShadow, 1 );	// PE0 = I/O strobe
	WrPortI ( IB0CR, NULL, 0x08 );		// permit write, chip select, 15 wait states
	fnMsDelay ( 1000 );						// wait for LCD to reset itself

	fn8Byte ( WCOMMAND, '\B00111100' );	// 8 bit mode
	fnUsDelay (100);
	fn8Byte ( WCOMMAND, '\B00111100' ); // 8 bit mode,  2 lines, 5x10
	fn8Byte ( WCOMMAND, '\B00001110' );	// turn on display and cursor: non-blinkin
	fn8Byte ( WCOMMAND, '\B00000110' );	// incr address and shift cursor with each character
}


/*********************************************************************
fn8Display - display a line of text on the LCD
	argument: address of null terminated text string
*/
void fn8Display ( char *szp )
{
	while ( *szp ) fn8Byte ( WDATA, *szp++ );
}

void fn8Clear ( void )
{
	fn8Byte ( WCOMMAND, 0x01 );	// clear the display
	fnMsDelay (3);						// insure at least 2 msec
}

void fn8Line2 ( void )
{
	fn8Byte ( WCOMMAND, 0xC0 );	// set RAM address to Line 2
					//note: this value may vary for different displays
	fnMsDelay (3);						// insure at least 2 msec
}

/********************************************************************/
/********************************************************************/
/********************************************************************/

void main ()
{
// in the following code, use the fn4... or fn8... as required by the interface

	fn8LCD_Init ();
		// IMPORTANT - the initialization sequence should only
		// be executed once after powerup.  Multiple executions
		// will cause the display to malfunction.
	
	while (1)
	{
		fn8Clear();
		fn8Display ( "0123456789" );
		fnMsDelay (500);
		fn8Clear();
		fn8Display ( "This is a test!" );
		fnMsDelay (500);

		fn8Line2 ();
		fn8Display ( "This is line TWO!" );
		fnMsDelay (500);
	}
}
